#include "LED.h"

// buffer size determines LED PWM frequency. LEDs signals are issued when generating audio signal.
// PWM frequency is audio signal frequency (44.1 kHz) divided by length of buffer. 
// Buffer size also determine duty cycle steps.
// here 44100 / 44 = 1002 Hz (~1kHz) and there are 44 possible duty cycle states.
// maximum buffer size is 255 because of using uint8_t index pointer
static uint8_t ledBuffer[44];
static uint8_t ledBufferIndex;

#define PWM_FIXED_ON 7
#define PWM_PERIOD ((44 + PWM_FIXED_ON) * 6)

/*
static const uint8_t ledBrightmessMap[] = {
	0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9,
	9, 10, 11, 11, 12, 13, 13, 14, 15, 16, 16, 17, 18, 19, 20, 20, 21,
	22, 23, 24, 25, 26, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
	38, 39, 40, 41, 42, 43, 44};
*/
	
static const uint8_t ledBrightmessMap[] = {
	0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5,
	6, 7, 7, 8, 9, 9, 10, 11, 12, 13, 13, 14, 15, 16, 17, 18, 18,
	19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
	38, 39, 40, 41, 42, 43, 44};


/*
static const uint8_t ledBrightmessMap[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
	9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
};
*/

// following array initialize random phase offsets and inc/dec direction
static uint8_t ledIntensity[LED_COUNT] = {0, 52, 25, 44, 29, 31, 0, 45};
static uint8_t currentDutyCycle[LED_COUNT] = {};
static uint8_t ledIntensityDirection[LED_COUNT] = {1, 1, 0, 1, 0, 0, 1, 0};
	
void LED_Init() {
	LED_PORT.DIR = 0xFF;
	
	// AVR128DB48 only
	// PORTMUX.TCDROUTEA = PORTMUX_TCD0_ALT1_gc;
	
	TCD0.CTRLA = TCD_CLKSEL_OSCHF_gc | TCD_CNTPRES_DIV4_gc;
	
	CCP = CCP_IOREG_gc;
	TCD0.FAULTCTRL = TCD_CMPAEN_bm | TCD_CMPBEN_bm;
	
	TCD0.DBGCTRL = TCD_DBGRUN_bm;
	TCD0.CMPASET = 0;
	TCD0.CMPACLR = 0;
	TCD0.CMPBSET = 0;
	TCD0.CMPBCLR = PWM_PERIOD + 1;
	
	while ((TCD0.STATUS & TCD_ENRDY_bm) == 0) {}
	TCD0.CTRLA = TCD_CLKSEL_OSCHF_gc | TCD_CNTPRES_DIV4_gc | TCD_ENABLE_bm;
}

void LED_Animate() {
	for (uint8_t led = 0; led < LED_COUNT; led++) {
		if (ledIntensityDirection[led]) {
			ledIntensity[led]++;
			if (ledIntensity[led] == sizeof(ledBrightmessMap) - 1) {
				ledIntensityDirection[led] = 0;
			}
		} else {
			ledIntensity[led]--;
			if (ledIntensity[led] == 0) {
				ledIntensityDirection[led] = 1;
			}
		}
		
		uint8_t newDutyCycle = ledBrightmessMap[ledIntensity[led]];
		if (newDutyCycle > currentDutyCycle[led]) {
			ledBuffer[newDutyCycle] |= (1 << led);
		} else if (newDutyCycle < currentDutyCycle[led]){
			ledBuffer[newDutyCycle] &= ~(1 << led);
		}
		currentDutyCycle[led] = newDutyCycle;
	}
	
	uint16_t newDutyCycleA = ledBrightmessMap[ledIntensity[6]] + PWM_FIXED_ON;
	uint16_t newDutyCycleB = ledBrightmessMap[ledIntensity[7]] + PWM_FIXED_ON;
	
	newDutyCycleA *= 6;
	newDutyCycleB *= 6;
	
		
	while ((TCD0.STATUS & TCD_ENRDY_bm) == 0) {}
	TCD0.CTRLA = TCD_CLKSEL_OSCHF_gc | TCD_CNTPRES_DIV4_gc;
	
	while ((TCD0.STATUS & TCD_CMDRDY_bm) == 0) {}
	TCD0.CMPACLR = newDutyCycleA;
	
	while ((TCD0.STATUS & TCD_CMDRDY_bm) == 0) {}
	TCD0.CMPBSET = PWM_PERIOD - newDutyCycleB;
	
	while ((TCD0.STATUS & TCD_ENRDY_bm) == 0) {}
	TCD0.CTRLA = TCD_CLKSEL_OSCHF_gc | TCD_CNTPRES_DIV4_gc | TCD_ENABLE_bm;
}

void LED_Show() {
	LED_PORT.OUT = ledBuffer[ledBufferIndex++] & 0xFF;
	//LED_PORT.OUT = 0xFF;
	if (ledBufferIndex == sizeof(ledBuffer)) {
		ledBufferIndex = 0;
	}
	
}