#ifndef W25Q_H_
#define W25Q_H_

#include <stddef.h>
#include <stdint.h>

#include "W25Q_Platform.h"

#define W25Q_REG_SR1_SRP0 (1 << 7)
#define W25Q_REG_SR1_SEC (1 << 6)
#define W25Q_REG_SR1_TB (1 << 5)
#define W25Q_REG_SR1_BP2 (1 << 4)
#define W25Q_REG_SR1_BP1 (1 << 3)
#define W25Q_REG_SR1_BP0 (1 << 2)
#define W25Q_REG_SR1_WEL (1 << 1)
#define W25Q_REG_SR1_BUSY (1 << 0)

#define W25Q_REG_SR2_SUS (1 << 7)
#define W25Q_REG_SR2_CMP (1 << 6)
#define W25Q_REG_SR2_LB3 (1 << 5)
#define W25Q_REG_SR2_LB2 (1 << 4)
#define W25Q_REG_SR2_LB1 (1 << 3)
#define W25Q_REG_SR2_QE (1 << 1)
#define W25Q_REG_SR2_SRP1 (1 << 0)

#define W25Q_REG_SR3_HOLD_RST (1 << 7)
#define W25Q_REG_SR3_DRV1 (1 << 6)
#define W25Q_REG_SR3_DRV0 (1 << 5)
#define W25Q_REG_SR3_WPS (1 << 2)

#define W25Q_OP_WriteSR1 0x01
#define W25Q_OP_PageProgram 0x02
#define W25Q_OP_ReadData 0x03
#define W25Q_OP_WriteDisable 0x04
#define W25Q_OP_ReadSR1 0x05
#define W25Q_OP_WriteEnable 0x06
#define W25Q_OP_FastRead 0x0B
#define W25Q_OP_BurstReadWithWrap 0x0C
#define W25Q_OP_WriteSR3 0x11
#define W25Q_OP_ReadSR3 0x15
#define W25Q_OP_SectorErase4K 0x20
#define W25Q_OP_WriteSR2 0x31
#define W25Q_OP_QuadPageProgram 0x32
#define W25Q_OP_ReadSR2 0x35
#define W25Q_OP_IndividualBlockLock 0x36
#define W25Q_OP_EnterQpiMode 0x38
#define W25Q_OP_IndividualBlockUnlock 0x39
#define W25Q_OP_FastReadDualOutput 0x3B
#define W25Q_OP_ReadBlockLock 0x3D
#define W25Q_OP_ProgramSecurityRegister 0x42
#define W25Q_OP_EraseSecurityRegister 0x44
#define W25Q_OP_ReadSecurityRegister 0x48
#define W25Q_OP_ReadUniqueId 0x4B
#define W25Q_OP_VolatileSRWriteEnable 0x50
#define W25Q_OP_BlockErase32K 0x52
#define W25Q_OP_ReadSfdpRegister 0x5A
#define W25Q_OP_ChipErase2 0x60
#define W25Q_OP_EnableReset 0x66
#define W25Q_OP_FastReadQuadOutput 0x6B
#define W25Q_OP_EraseSuspend 0x75
#define W25Q_OP_ProgramSuspend 0x75
#define W25Q_OP_SetBurstWithWrap 0x77
#define W25Q_OP_EraseResume 0x7A
#define W25Q_OP_ProgramResume 0x7A
#define W25Q_OP_GlobalBlockLock 0x7E
#define W25Q_OP_DeviceId 0x90
#define W25Q_OP_MftrDeviceIdDualIO 0x92
#define W25Q_OP_MftrDeviceIdQuadIO 0x94
#define W25Q_OP_GlobalBlockUnlock 0x98
#define W25Q_OP_Reset 0x99
#define W25Q_OP_JedecId 0x9F
#define W25Q_OP_ReleasePowerDown 0xAB
#define W25Q_OP_Id 0xAB
#define W25Q_OP_PowerDown 0xB9
#define W25Q_OP_FastReadDualIO 0xBB
#define W25Q_OP_SetReadParameters 0xC0
#define W25Q_OP_ChipErase 0xC7
#define W25Q_OP_BlockErase64K 0xD8
#define W25Q_OP_OctalWordReadQuadIO 0xE3
#define W25Q_OP_WordReadQuadIO 0xE7
#define W25Q_OP_FastReadQuadIO 0xEB
#define W25Q_OP_ExitQPIMode 0xFF

typedef struct {
	uint8_t parameterIdLsb;
	uint8_t minorRevision;
	uint8_t majorRevision;
	uint8_t length;

	uint8_t ptrBytes[3];
	uint8_t parameterIdMsb;
} SfdpParameterHeader;

typedef struct {
	uint32_t signature;

	uint8_t minorRevision;
	uint8_t majorRevision;
	uint8_t parametersCount;
	uint8_t accessProtocol;

	SfdpParameterHeader parameters[];
} SfdpHeader;

typedef struct {
	// 1st double word
	uint8_t sectorEraseSizes :2;
	uint8_t writeGranuality :1;
	uint8_t volatileStatusRegisterBlockProtectBits :1;
	uint8_t writeEnableInstructionSelectForWritingToVolatileStatusRegister :1;
	uint8_t unused111 :3;
	uint8_t erase4KbInstruction :8;
	uint8_t Support1_1_2FastRead :1;
	uint8_t AddressBytes :2;
	uint8_t SupportsDoubleTransferRate :1;
	uint8_t Support1_2_2FastRead :1;
	uint8_t Support1_4_4FastRead :1;
	uint8_t Support1_1_4FastRead :1;
	uint16_t unused1FF :9;

	// 2nd double word
	uint32_t density;

	// 3rd double word
	uint8_t fastRead1_4_4NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead1_4_4NumberOfModeClocks :3;
	uint8_t fastRead1_4_4Instruction :8;
	uint8_t fastRead1_1_4NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead1_1_4NumberOfModeClocks :3;
	uint8_t fastRead1_1_4Instruction :8;

	// 4th double word
	uint8_t fastRead1_1_2NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead1_1_2NumberOfModeClocks :3;
	uint8_t fastRead1_1_2Instruction :8;
	uint8_t fastRead1_2_2NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead1_2_2NumberOfModeClocks :3;
	uint8_t fastRead1_2_2Instruction :8;

	// 5th double word
	uint8_t supports2_2_2FastRead :1;
	uint8_t reserved111_1 :3;
	uint8_t supports4_4_4FastRead :1;
	uint32_t reserved111111111111111111111111111 :27;

	// 6th double word
	uint16_t reservedFFFF;
	uint8_t fastRead2_2_2NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead2_2_2NumberOfModeClocks :3;
	uint8_t fastRead2_2_2Instruction :8;

	// 7th double word
	uint16_t reservedFFFF_1;
	uint8_t fastRead4_4_4NumberOfWaitStatesNeededBeforeValidOutput :5;
	uint8_t fastRead4_4_4NumberOfModeClocks :3;
	uint8_t fastRead4_4_4Instruction :8;

	// 8th double word
	uint8_t eraseType1Size;
	uint8_t eraseType1Instruction;
	uint8_t eraseType2Size;
	uint8_t eraseType2Instruction;

	// 9th double word
	uint8_t eraseType3Size;
	uint8_t eraseType3Instruction;
	uint8_t eraseType4Size;
	uint8_t eraseType4Instruction;

} SfdpBasicFlashParameters;

#define W25Q_PROTECTION_MODE_AREA_LOCK 0
#define W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK 1

#define W25Q_DRIVER_STRENGTH_100 0x00
#define W25Q_DRIVER_STRENGTH_75 0x01
#define W25Q_DRIVER_STRENGTH_50 0x02
#define W25Q_DRIVER_STRENGTH_25 0x03

#define W25Q_HOLD_RESET_MODE_HOLD 0
#define W25Q_HOLD_RESET_MODE_RESET 1
#define W25Q_HOLD_RESET_MODE_IO 2

#define W25Q_SECURITY_REGISTER_1_ADDRESS 0x00001000
#define W25Q_SECURITY_REGISTER_2_ADDRESS 0x00002000
#define W25Q_SECURITY_REGISTER_3_ADDRESS 0x00003000

#define W25Q_STATUS_REGISTER_PROTECTION_SOFTWARE 0
#define W25Q_STATUS_REGISTER_PROTECTION_HARDWARE 1
#define W25Q_STATUS_REGISTER_PROTECTION_POWER_SUPLY_LOCKDOWN 2
#define W25Q_STATUS_REGISTER_PROTECTION_ONE_TIME_PROGRAM 3

/**
 * Initializes SPI bus to the flash memory. It must be called before using library.
 */
void W25Q_Init();

/**
 * Executes Write Enable instruction against flash memory. It must be called before any write/erase
 * operation and high levels calls of this library usually calls it autmatically when needed.
 */
void W25Q_WriteEnable();

/**
 * Enables writing into status register in volatile mode which means that settings applied to status
 * register wont be persistent across power loss or reset.
 */
void W25Q_WriteStatusRegisterVolatileEnable();

/**
 * Disables writing which was enabled using W25Q_WriteEnable function.
 */
void W25Q_WriteDisable();

/**
 * Reads 1st status register from flash memory.
 */
uint8_t W25Q_ReadSR1();

/**
 * Reads 2nd status register from flash memory.
 */
uint8_t W25Q_ReadSR2();

/**
 * Reads 3rd status register from flash memory.
 */
uint8_t W25Q_ReadSR3();

/**
 * Updates 1st status register in flash memory. Change made in non-volatile register will be persistent.
 * If you want it do it using volatile way call W25Q_WriteSR1Volatile instead.
 */
void W25Q_WriteSR1(uint8_t newRegValue);

/**
 * Updates 2nd status register in flash memory. Change made in non-volatile register will be persistent.
 * If you want it do it using volatile way call W25Q_WriteSR5Volatile instead.
 */
void W25Q_WriteSR2(uint8_t newRegValue);

/**
 * Updates 3rd status register in flash memory. Change made in non-volatile register will be persistent.
 * If you want it do it using volatile way call W25Q_WriteSR3Volatile instead.
 */
void W25Q_WriteSR3(uint8_t newRegValue);

/**
 * Updates 1st status register in flash memory. Changes applied to register is volatile. That means that
 * after power loss or reset register will be restored to original value, not to value changed using this
 * call. If you want to change value in non-volatile manner use function W25Q_WriteSR1 instead.
 */
void W25Q_WriteSR1Volatile(uint8_t newRegValue);

/**
 * Updates 2nd status register in flash memory. Changes applied to register is volatile. That means that
 * after power loss or reset register will be restored to original value, not to value changed using this
 * call. If you want to change value in non-volatile manner use function W25Q_WriteSR2 instead.
 */
void W25Q_WriteSR2Volatile(uint8_t newRegValue);

/**
 * Updates 3rd status register in flash memory. Changes applied to register is volatile. That means that
 * after power loss or reset register will be restored to original value, not to value changed using this
 * call. If you want to change value in non-volatile manner use function W25Q_WriteSR3 instead.
 */
void W25Q_WriteSR3Volatile(uint8_t newRegValue);

/**
 * Reads data from memory at specified address. Address does not require any alignment. count of read bytes
 * could by any number less than memory capacity. Memory capacity could be read using W25Q_GetCapacity function.
 * This function working only when CLK frequncy is less or equal than 33 Mhz. If you want to read data full
 * speed (ie 104 Mhz) and your flash memory supports it use W25Q_ReadDataFast instead. It has the same interface.
 *
 * Returns 0 on success any other value on error.
 */
uint8_t W25Q_ReadData(uint32_t address, uint8_t* output, uint16_t outputLength);

/**
 * Reads data from memory at specified address. Address does not require any alignment. count of read bytes
 * could by any number less than memory capacity. Memory capacity could be read using W25Q_GetCapacity function.
 * This function working only when CLK frequncy is less or equal than 33 Mhz. If you want to read data full
 * speed (ie 104 Mhz) and your flash memory supports it use W25Q_ReadDataFast instead. It has the same interface.
 *
 * Returns 0 on success any other value on error.
 */
uint8_t W25Q_ReadDataIrq(uint32_t address, uint8_t* output, uint16_t outputLength);

/**
 * Reads data from memory at specified address and supports reading at full frequency. Cunt od read bytes could
 * be any number less than memory capacity. Memory capacity could be read using W25Q_GetCapacity function.
 *
 * Returns 0 on success any other value on error.
 */
uint8_t W25Q_ReadDataFast(uint32_t address, uint8_t* output, uint16_t outputLength);

/**
 * Verifies if flash memory contains specified data at specified address. Returns 0 if data are the same and 1
 * if there are some differences betwwen data in falsh and verified data.
 */
uint8_t W25Q_VerifyData(uint32_t address, uint8_t* data, uint16_t dataLength);

/**
 * Programs one page of data. Page is 256 bytes long. The address do not need to be aligned. If writing
 * overflows from range of page it will wrap to begin of the programmed page (not the next page!). If there
 * are more than 256 bytes there will overwrite previously written data.
 *
 * Returns 0 on success any other value means error.
 */
uint8_t W25Q_ProgramPage(uint32_t address, uint8_t* data, uint16_t dataLenth);

/**
 * Erases sector 4K which contains specified address. Address do not need to be aligned.
 *
 * Returns 0 on success any other value means error.
 */
uint8_t W25Q_EraseSector(uint32_t address);

/**
 * Erase entire chip. Returns 0 on success.
 */
uint8_t W25Q_EraseChip();

/**
 * Reads manufacturer ID and device ID from lash memory.
 */
void W25Q_ReadManufacturerId(uint8_t* vendorIdOutput, uint8_t* deviceIdOutput);

/**
 * Returns unique ID of device.
 */
uint64_t W25Q_ReadUniqueId();

/**
 * Returns ID based on JEDEC standard.
 */
uint32_t W25Q_ReadJedecId();

/**
 * Reads content of specified SFDP parameter. Capacity of output must be greather or equal then
 * param->length * 4.
 */
void W25Q_ReadSfdpParam(SfdpParameterHeader* param, uint8_t* output);

/**
 * Returns W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK or W25Q_PROTECTION_MODE_AREA_LOCK based on
 * current protection mode used. You can change that settings using W25Q_SetCurrentProtectionMode function.
 */
uint8_t W25Q_GetCurrentProtectionMode();

/**
 * Sets protection mode. First parameter is W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK or
 * W25Q_PROTECTION_MODE_AREA_LOCK and second parameter specifies if the settings will persist reset
 * or it wont. You can read current mode using W25Q_GetCurrentProtectionMode function. Returns 0 on success
 * and any other value on error.
 */
uint8_t W25Q_SetCurrentProtectionMode(uint8_t newMode, uint8_t writeVolatile);

/**
 * Gets one of W25Q_DRIVER_STRENGTH_100, W25Q_DRIVER_STRENGTH_75, W25Q_DRIVER_STRENGTH_50 and
 * W25Q_DRIVER_STRENGTH_25 based on current driver strength settings. You can change that setting
 * using W25Q_SetDriverStrength function.
 */
uint8_t W25Q_GetDriverStrength();

/**
 * Sets driver strength. First parameter must be one of W25Q_DRIVER_STRENGTH_100,
 * W25Q_DRIVER_STRENGTH_75, W25Q_DRIVER_STRENGTH_50 or  W25Q_DRIVER_STRENGTH_25. Second parameter
 * specifies if the settings will persist reset or it wont. You can read current status using
 * W25Q_GetDriverStrength function. Returns 0 on success and any other value on error.
 */
uint8_t W25Q_SetDriverStrength(uint8_t newStrength, uint8_t writeVolatile);

/**
 * Returns if HOLD/RESET pin is in HOLD or RESET mode. Output value is one of W25Q_HOLD_RESET_MODE_RESET,
 * W25Q_HOLD_RESET_MODE_HOLD and W25Q_HOLD_RESET_MODE_IO. You can change settings using
 * W25Q_SetHoldResetPinMode function.
 */
uint8_t W25Q_GetHoldResetPinMode();

/**
 * Sets HOLD/RESET pin to specified function. First parameter must be W25Q_HOLD_RESET_MODE_RESET and
 * W25Q_HOLD_RESET_MODE_HOLD. Second parameter specifies if the settings will persist reset or it wont.
 * You can read current settings using W25Q_GetHoldResetPinMode function. If you want to set pin into IO
 * mode for using quad data transfer use W25Q_EnableQuadMode(). function Returns 0 on success and any
 * other value on error.
 */
uint8_t W25Q_SetHoldResetPinMode(uint8_t newMode, uint8_t writeVolatile);

/**
 * Returns capacity of connected flash memory. It works only for flashes which supports JEDEC standard.
 * Output in bytes is stored into passed pointer. Returns 0 on success and any other value on error.
 */
uint8_t W25Q_GetCapacity(uint64_t* output);

/**
 * Programs specified security register. Address don't need to be aligned. Security registers has capacity
 * of 256 bytes and write cannot exceed this capacity. Use W25Q_SECURITY_REGISTER_x_ADDRESS (x is number
 * of register) for addressing registers. For example writing to 10th byte of 2nd register could be done
 * using call W25Q_ProgramSecurityRegister(W25Q_SECURITY_REGISTER_2_ADDRESS + 9, ???, ???);
 *
 * Returns 0 on success any other value on error.
 */
uint8_t W25Q_ProgramSecurityRegister(uint32_t address, uint8_t* data, uint16_t length);

/**
 * Erases content of security register which contains specified address.
 */
void W25Q_EraseSecurityRegister(uint32_t address);

/**
 * Reads content of security register. Address don't need to be aligned. Security registers has capacity
 * of 256 bytes and reading cannot exceed this capacity. Use W25Q_SECURITY_REGISTER_x_ADDRESS (x is number
 * of register) for addressing registers.
 */
uint8_t W25Q_ReadSecurityRegister(uint32_t address, uint8_t* output, uint16_t outputLength);

/**
 * Enters using QPI mode. Remember that this library cannot work with QPI mode because it is using SPI
 * peripheral of microcontroller.
 */
void W25Q_EnterQpiMode();

/**
 * Sets device into power down mode. To use device when power down you must release it from power down
 * mode using W25Q_ReleasePowerDown function before.
 */
void W25Q_PowerDown();

/**
 * Release device from power down mode.
 */
void W25Q_ReleasePowerDown();

/**
 * Locks block containing specified address. Locking block works only when current memory protection mode
 * is W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK. Blocks are 64KB large.
 */
void W25Q_LockBlock(uint32_t address);

/**
 * Unlocks block containing specified address. Locking block works only when current memory protection mode
 * is W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK. Blocks are 64KB large.
 */
void W25Q_UnlockBlock(uint32_t address);

/**
 * Returns 1 if block containing specified address is locked. Locking block works only when current memory
 * protection mode is W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK. Blocks are 64KB large.
 */
uint8_t W25Q_IsBlockLocked(uint32_t address);

/**
 * Applies lock to all blocks. Locking block works only when current memory protection mode is
 * W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK.
 */
void W25Q_LockAllBLocks();

/**
 * Removes lock from all blocks. Locking block works only when current memory protection mode is
 * W25Q_PROTECTION_MODE_INDIVIDUAL_BLOCK_LOCK.
 */
void W25Q_UnlockAllBlocks();

/**
 * Resets device.
 */
void W25Q_Reset();

/**
 * Returns array lock settings. Arary lock settings works only when current memory protection mode is
 * W25Q_PROTECTION_MODE_AREA_LOCK.
 */
void W25Q_GetArrayLock(uint8_t* sec, uint8_t* tb, uint8_t* bp, uint8_t* cmp);

/**
 * Sets array lock settings. Arary lock settings works only when current memory protection mode is
 * W25Q_PROTECTION_MODE_AREA_LOCK.
 */
void W25Q_SetArrayLock(uint8_t sec, uint8_t tb, uint8_t bp, uint8_t cmp, uint8_t writeVolatile);

/**
 * Gets current status register protection mode. Returns one of W25Q_STATUS_REGISTER_PROTECTION_SOFTWARE,
 * W25Q_STATUS_REGISTER_PROTECTION_HARDWARE, W25Q_STATUS_REGISTER_PROTECTION_POWER_SUPLY_LOCKDOWN and
 * W25Q_STATUS_REGISTER_PROTECTION_ONE_TIME_PROGRAM. You can set this settings using
 * W25Q_SetStatusRegisterProtection function.
 */
uint8_t W25Q_GetStatusRegisterProtection();

/**
 * Sets status register protection mode. First parameter must be one of
 * W25Q_STATUS_REGISTER_PROTECTION_SOFTWARE, W25Q_STATUS_REGISTER_PROTECTION_HARDWARE,
 * W25Q_STATUS_REGISTER_PROTECTION_POWER_SUPLY_LOCKDOWN and
 * W25Q_STATUS_REGISTER_PROTECTION_ONE_TIME_PROGRAM. Second parameter specifies if the settings will
 * persist reset or it wont. You can read current protection mode using W25Q_GetStatusRegisterProtection
 * function. Returns 0 on success and any other value on error.
 */
uint8_t W25Q_SetStatusRegisterProtection(uint8_t newProtection, uint8_t writeVolatile);

/**
 * Enables quad mode so you can then transfer data in paralel and increase bandwidth. Enabling quad mode
 * disables WP and HOLD/RESET pins.
 */
uint8_t W25Q_EnableQuadMode(uint8_t writeVolatile);

/**
 * Disables quad mode. After issuing this you cannot transafer data in paralel using 4 lines but pins WP
 * and HOLD/RESET could be used in they native form.
 */
uint8_t W25Q_DisableQuadMode(uint8_t writeVolatile);

/**
 * Returns information if quad mode is enabled.
 */
uint8_t W25Q_IsQuadModeEnabled();

/**
 * Makes security register locked and read-only forever. THIS CANNOT BE UNDONE. First parameter is number
 * of register to be locked. Remember that register are indexed by 1. So valid values are 1, 2 or 3.
 */
uint8_t W25Q_LockSecurityRegister(uint8_t registerNum);

/**
 * Check if specified security register is locked. Remember that register are indexed by 1. So valid
 * values are 1, 2 or 3.
 */
uint8_t W25Q_IsSecurityRegisterLocked(uint8_t registerNum);

#endif
